function tuparam(varargin)  % version 8/28/2001
% TUPARAM: EXTRACT TUBE MODEL PARAMETERS.
% Copyright (C) 2001 by Norman Koren. The author assumes no legal liability arising
% from the use of this program. All the usual disclaimers apply.
%
% Uses fminsearch (fmins in older Matlab versions). Calls optube, tucalc.
% NOPT is the number of parameters to optimize. Generally 5.
% Setting NOPT to 4 eliminates KVB.
%
% A good initial guess for MU helps convergence a lot.
% A good initial guess for KVB (knee voltage) helps for Pentodes.
% A global minimum is reached when solution looks good
% and doesn't depend much on initial guess.

%%%%%%%%%%%%%  SET THE TUBE TYPE HERE %%%%%%%%%%%%%%%

if (nargin==1)
   filename = varargin{1};  % data file (convert from cell.)
else
   error(['Need one input argument (the data file): ' num2str(nargin)]);
end

global TubeType MU EX KG1 KGX KP KVB Vp Vg Vs Vsi Idata ...
   ERR VCT PENFN WEIGHT NOPT SOLVER

% Initial parameter values for start of optimization.
% Better values can be entered with tube data if needed.
% fmins requires better starting values than fsolve (Optim. toolbox).
MU = 20;	   	% mult. x(1).  Amplification factor.
EX = 1.35;		% mult. x(2).  Exponent.
KG1 = 1000;    % Use KGX for KG1 inside optimization.
KP = 200;	 	% mult. x(4).
KVB = [];		% mult. x(5).  Knee volts:  mostly pentode.
KG2 = 4500;    % Inverse grid current sensitivity: NOT optimized. Entered value kept.

% The following are default parameters for control variables, such as
% plot sweep variables.  Reset if needed in tube data.
WEIGHT = 2;		% 1: Optimize difference in currents.
		      	% 2: logarithmic difference; 3: relative differences.
Vpinc = 5;		% Plate V increment.
Vpmax = 450;	% Maximum plate voltage.
VGP = 0:-5:-30;	% Default grid V for plot.
VGTRI = [];		% Grid V for pentode triode-mode plot if entered.
Vs = [];		% Screen voltage (Enter value(s) for pentode only).
axset = [];		% Enter [0 xmax 0 ymax] for manual axis.
VCT = 0;	   	% Contact potential: usually 0. Always added to Vg in equations.
%	VCT is ~.5V for 12AX7, 12AU7 in old RCA, Sylvania manuals.
SOLVER = 'fmins';  % Set SOLVER = 'fminsearch' for ver. 5.3+; 'fmins' for earlier
% SOLVER = 'fsolve';  % for Optimization toolbox.
PENFN = 1;		% For Pentodes only: 1 for atan; 2 for (1-exp); 3 for tanh.
penf = ['(atan) ';'(1-exp)';'(tanh) '];
NOPT = 5;      % No. of parameters to optimize. Set later.
% versions. If the optimization toolbox is available, use fsolve.

% The following is data from tube plate curves.
% Enter Source, TubeName, arrays Vp, Vg, Idata, and any plot
% parameters different from the above defaults.
% For pentodes, enter Vs.

%                              READ THE TUBE DATA FILE.

if (isempty(findstr(filename,'.')))  % Make .m the default extension.
   filename = [filename '.m'];  end
[fct,MESSAGE] = fopen(filename, 'rt');  % Read the control file.
if (fct<0) disp([MESSAGE ' ' filename])
   disp('Directory of m-files:');  dir *.m,  return;  end
while 1  line = fgetl(fct);  if ~isstr(line),  break,  end
   eval(line,'fprintf(''%s%s\n'',7,lasterr);');  end;  fclose(fct);
[fct,MESSAGE] = fopen(filename, 'rt');  % Reread the control file.
while 1  line = fgetl(fct);  if ~isstr(line),  break,  end
   eval(line,'fprintf(''%s%s\n'',7,lasterr);');  end;  fclose(fct);
   
%                        End input from filename (control file)

Vpi = Vp;	% Save initial data.
vplth = length(Vp);
KGX = KG1^(1/EX);   % Optimize KGX. Works better than KG1.

% Expand Vs if entered as a scalar.
if (~isempty(Vs))
   TubeType = 'PENTODE';
   if (length(Vs)==1) Vs = Vs.*ones(size(Vp));
   elseif (length(Vs)~=vplth)
      error('Vs length must be 1 or same as Vp.');  end
Vsi = Vs;  % Save initial data.
else
   TubeType = 'TRIODE';
   % if (max(Vg+VCT)<eps) NOPT = 4;  end  % No. of parameters to optimize.
end

if (isempty(KVB))		% Guess KVB for pentode or triode.
   if (strcmp(TubeType(1:4),'PENT')) KVB = 40;
   else KVB = 300;	 end	% Guess for small signal triode. (Was 15).
end

if (length(Vg) ~= vplth)
   error('Length of Vg differs from Vp.');  end
if (length(Idata) ~= vplth)
   error('Length of Idata differs from Vp.');  end

% x(i) is multiplied by MU, EX, KGX, KP, KVB.

%                        DO OPTIMIZATION.  Enter arrays.

runstr = ['Running ',TubeName,'...     '];
fprintf(1,runstr);
x = ones(1,5);  x0 = x(1:NOPT);

if (strcmp(SOLVER,'fminsearch'))  % Matlab 5.3 or later
   OPTIONS = optimset('fminsearch');      % (new)
   OPTIONS = optimset('MaxFunEvals',5000);
   x = fminsearch('OpTube',x0,OPTIONS);	% Standard function (new)
elseif (strcmp(SOLVER,'fmins'))  % Matlab 5.2 or earlier.
   % OPTIONS = FOPTIONS;  
   OPTIONS(2:3) = 1e-5*ones(1,2);  % Termination Tol. X, F.
   OPTIONS(14) = 5000;  % Max evaluations.
   x = fmins('OpTube',x0,OPTIONS);	% Standard function: Old Matlab versions.
elseif (strcmp(SOLVER,'fsolve'))
   x = fsolve('OpTube',x0);	% Optimization Toolbox, if available.
end  %                       END OPTIMIZATION
exx = EX;
if (strcmp(TubeType,'TRIODE') | (max(Vs)-min(Vs))/max(Vs)>.3)
   exx = EX*x(2);  end  % Use optimization on EX.

% Typical output formats (Print first 2 lines):

% .SUBCKT 6550 1 2 3 4 ;  P G1 C G2
% + PARAMS: MU=7.9 EX=1.35 KG1=890 KG2=4200 KP=60 KVB=24
% +         CCG=14P CPG1=.85P CCP=12P RGI=1K 
% RE1  7 0  10MEG    ; DUMMY SO NODE 7 HAS 2 CONNECTIONS
% E1   7 0  VALUE=  ; E1 BREAKS UP LONG EQUATION FOR G1.
% +{V(4,3)/KP*LOG(1+EXP((1/MU+V(2,3)/V(4,3))*KP))}
% G1   1 3  VALUE={(PWR(V(7),EX)+PWRS(V(7),EX))/KG1*ATAN(V(1,3)/KVB)}
% G2   4 3  VALUE={(EXP(EX*(LOG((V(4,3)/MU)+V(2,3)))))/KG2}

% .SUBCKT 12AU7 1 2 3  ; P G C; NEW MODEL
% + PARAMS: MU=21.5 EX=1.3 KG1=1180 KP=84 KVB=300 RGI=2000
% + CCG=2.3P  CGP=2.2P CCP=1.0P  ; ADD .7PF TO ADJACENT PINS; .5 TO OTHERS. 
% E1 7 0 VALUE=
% +{V(1,3)/KP*LOG(1+EXP(KP*(1/MU+V(2,3)/SQRT(KVB+V(1,3)*V(1,3)))))}
% RE1 7 0 1G
% G1 1 3 VALUE={(PWR(V(7),EX)+PWRS(V(7),EX))/KG1}


fprintf(1,'     ERR =%7.3g   WEIGHT = %d\n\n',ERR,WEIGHT);
Outlin1 = ['.SUBCKT ' TubeName];
Outlin2 = sprintf('MU=%6.2f  EX=%6.3f  KG1=%6.1f  KP=%6.2f ', ...
   [MU*x(1) exx (KGX*x(3))^exx KP*x(4)]);
Outlin3 = '';
if     (NOPT>=5 & strcmp(TubeType,'TRIODE') & (max(Vg)+VCT)>eps)
     Outlin3 = [sprintf('KVB=%6.1f',KVB*(x(5))^2)];
elseif (NOPT>=5 & strcmp(TubeType,'PENTODE'))
     Outlin3 = [sprintf('KVB=%6.1f',KVB*x(5))];
else Outlin3 = [sprintf('KVB=%6.1f',KVB)];  end;
  
Outlin3 = [Outlin3 sprintf('  VCT=%6.2f',VCT)];
if (strcmp(TubeType,'TRIODE'))
   fprintf(1,'.SUBCKT %s  1 2 3 ; P G C (Triode)\n',TubeName);
   Outlin1 = [Outlin1,' ;  Triode'];
elseif (strcmp(TubeType,'PENTODE'))
   fprintf(1,'.SUBCKT %s  1 2 3 4 ; P G1 C G2 (Pentode)\n',TubeName);
   Outlin1 = [Outlin1,' ;  Pentode'];
   Outlin2 = [Outlin2 '  KG2= ' num2str(KG2)];
end

fprintf(1,['+ PARAMS: ',Outlin2,'\n'])
if (length(Outlin3)) fprintf(1,['+ ',Outlin3,'\n']); end
fprintf(1,'* %s  %s\n.ENDS\n\n',Source, date);

if (strcmp(TubeType,'PENTODE2'))	% Print .FUNC command.
   if     (PENFN==1) fprintf(1,'.FUNC PENFN(X) {ATAN(X)}\n\n');
   elseif (PENFN==2) fprintf(1,'.FUNC PENFN(X) {1-EXP(-X)}\n\n')
   elseif (PENFN==3) fprintf(1, ...
    '.FUNC PENFN(X) {(EXP(X)-EXP(-X))/(EXP(X)+EXP(-X))} ; TANH.\n\n')
end, end

Vp = 0:Vpinc:Vpmax;
if (strcmp(TubeType(1:4),'PENT'))  Vs = Vsi(1).*ones(size(Vp));  end  % For plot.
   
ip = zeros(length(VGP),length(Vp));
for (g = 1:length(VGP))
   Vg = VGP(g);  ip(g,:) = tucalc(x);
end   
figure;  plot(Vp, ip); grid on;
if (length(axset)==4) axis(axset); end
axsav = axis;
xlabel('Plate volts');  ylabel ('Plate current (A)');
title([TubeName,' PLATE CHARACTERISTICS  ',date]);
% Plot original data points.
hold on
if (strcmp(TubeType,'TRIODE'))
   plot(Vpi, Idata, 'o')
else	% Plot 'x' oentode curve; 'o' on triode.
   for (k = 1:length(Vpi))
     if (Vsi(k) == Vsi(1)) plot(Vpi(k), Idata(k), 'x')
     else plot(Vpi(k), Idata(k), 'o'),  end
   end
end
hold off
y1 = max(max(ip));  x1 = max(Vp);
ya = .95*axsav(4);  yb = .05*axsav(4);  xa = .04*axsav(2);
text(xa, ya, ['Vg = ',sprintf('%5.1f ',VGP)]), ya = ya-yb;
text(xa, ya, ['* ',Source]), ya = ya-yb;
text(xa, ya, Outlin1), ya = ya-yb;
text(xa, ya, Outlin2), ya = ya-yb;
if (length(Outlin3))
   text(xa, ya, Outlin3), ya = ya-yb; end

%  Plot TRIODE data for pentodes.
if (strcmp(TubeType(1:4),'PENT'))
   text(xa, ya, sprintf('Vs for pentode curves = %6.1f',Vsi(1)))
   ya = ya-yb;
   hold on
   if (length(VGTRI)) Vgplt = VGTRI;
      text(xa, ya, ['Vg (triode) = ',sprintf('%5.1f ',Vgplt)]);
      ya = ya-yb;
   else Vgplt = VGP; end
   Vp = Vpinc:Vpinc:Vpmax;
   vplth = length(Vp);
   ip = zeros(length(Vg),vplth);
   Vs = Vp;
   for (g = 1:length(Vgplt))
      Vg = Vgplt(g);  ip(g,:) = tucalc(x);
   end   
   plot(Vp, ip, '--'); grid on;
   axis(axsav);
   hold off
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Here is the format for the Pspice .LIB file:

% .SUBCKT 12AU7 P G C  ; (triode)
% XV1 P G C 12AX7
% + PARAMS: MU= 22.98 EX= 1.295 KG1= 213.8 KP= 79.94 
% + KVB=  10.0 VCT=  0.00
% * Tom Mitchell  30-Nov-97
% .ENDS
